/**
 * <copyright>
 * </copyright>
 *
 * $Id: LayoutData.java 9037 2012-01-27 17:15:03Z laslovd $
 */
package org.openxma.dsl.pom.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.xtext.EcoreUtil2;
import org.openxma.dsl.core.model.Style;
import org.openxma.dsl.pom.PomFactory;
import org.openxma.dsl.pom.PomPackage;
import org.openxma.dsl.pom.model.AttachmentOwnPosition;
import org.openxma.dsl.pom.model.AttachmentProperty;
import org.openxma.dsl.pom.model.AttachmentSpecification;
import org.openxma.dsl.pom.model.ComposeData;
import org.openxma.dsl.pom.model.Composite;
import org.openxma.dsl.pom.model.GuiElement;
import org.openxma.dsl.pom.model.HeightProperty;
import org.openxma.dsl.pom.model.IComposite;
import org.openxma.dsl.pom.model.IElementWithLayoutData;
import org.openxma.dsl.pom.model.LayoutDataProperty;
import org.openxma.dsl.pom.model.SetOfGuiElements;
import org.openxma.dsl.pom.model.StyleProperty;
import org.openxma.dsl.pom.model.WidthProperty;


public class LayoutData {
	
	private final IElementWithLayoutData elementWithLayoutData;
	
	/**
	 * Constructor
	 * 
	 * @param elementWithLayoutData
	 */
	public LayoutData(IElementWithLayoutData elementWithLayoutData) {
		if (elementWithLayoutData == null) {
			throw new IllegalArgumentException("The elementWithLayoutData must not be null");
		}
		this.elementWithLayoutData = elementWithLayoutData;
	}
	
	public IElementWithLayoutData getElementWithLayoutData() {
		return elementWithLayoutData;
	}
	
	/**
	 * Returns the value of the '<em><b>Properties</b></em>' containment reference list.
	 * The list contents are of type {@link org.openxma.dsl.pom.model.LayoutDataProperty}.
	 * <!-- begin-user-doc -->
	 * <p>
	 * If the meaning of the '<em>Properties</em>' containment reference list isn't clear,
	 * there really should be more of a description here...
	 * </p>
	 * <!-- end-user-doc -->
	 * @return the value of the '<em>Properties</em>' containment reference list.
	 * @see org.openxma.dsl.pom.PomPackage#getLayoutData_Properties()
	 * @model containment="true"
	 * @generated
	 */
	public List<LayoutDataProperty> getProperties() {
		return elementWithLayoutData.getLayoutProperties();
	}
	/**
	 * Get the xmadsl attachment for the specified own position. In case no has
	 * been found for left or top try to find center attachments: .) hcenter for
	 * left .) vcenter for top
	 * 
	 * (In XMA the center attachments can be attached to the own top or left
	 * attachment.)
	 * 
	 * @param layoutData
	 * @param ownPosition
	 * @modified
	 * @return
	 */
	public static AttachmentSpecification getAttachment(LayoutData layoutData, final AttachmentOwnPosition ownPosition) {
		
		AttachmentSpecification as = getAttachment(getAttachments(layoutData),ownPosition);	
		if (as != null) return as;
		
		// Try center attachments
		switch (ownPosition.getValue()) {
		case AttachmentOwnPosition.LEFT_VALUE:
			return getAttachment(layoutData, AttachmentOwnPosition.HCENTER);
		case AttachmentOwnPosition.TOP_VALUE:
			return getAttachment(layoutData, AttachmentOwnPosition.VCENTER);
		}
		return null;
	}
	
	public static AttachmentSpecification getAttachment(List<AttachmentProperty> attachments, final AttachmentOwnPosition ownPosition) {
		for (AttachmentProperty attachment : attachments) {
			if (attachment.getOwnPosition().equals(ownPosition))
				return attachment.getAttachmentSpecification();
		}
		return null;
	}
	
	
	
	/**
	 *
	 * @modified 
	 */
	public static List<AttachmentProperty> getAttachments(Collection<LayoutDataProperty> layoutDataProperties) {
		List<AttachmentProperty> attachments = new ArrayList<AttachmentProperty>();
		for (LayoutDataProperty layoutDataProperty : layoutDataProperties) {
			if (layoutDataProperty instanceof AttachmentProperty) {
				attachments.add((AttachmentProperty) layoutDataProperty);
			}
		}
		return attachments;
	}		
	
	/**
	 *
	 * @modified 
	 */
	public List<AttachmentProperty> getAttachments() {
		return getAttachments(getProperties());
	}	

	/**
	 * @param layoutData
	 * @modified
	 */
	private static List<AttachmentProperty> getAttachments(LayoutData layoutData) {			
		if (null != layoutData) {
			return layoutData.getAttachments();
		} else {
			return Collections.emptyList();	
		}
	}

	/**
	 * @modified
	 * @return the top attachment
	 */
	public AttachmentSpecification getTopAttachment() {
		return getAttachment(this, AttachmentOwnPosition.TOP);
	}

	/**
	 * @modified
	 * @return the left attachment
	 */
	public AttachmentSpecification getLeftAttachment() {
		return getAttachment(this, AttachmentOwnPosition.LEFT);
	}

	/**
	 * @modified
	 * @return the right attachment
	 */
	public AttachmentSpecification getRightAttachment() {
		return getAttachment(this, AttachmentOwnPosition.RIGHT);
	}

	/**
	 * @modified
	 * @return the bottom attachment
	 */
	public AttachmentSpecification getBottomAttachment() {
		return getAttachment(this, AttachmentOwnPosition.BOTTOM);
	}

	/**
	 * 
	 * @modified
	 * @return
	 */
	public ComposeData getComposeData() {
		GuiElement guiElement = (GuiElement) elementWithLayoutData.eContainer();
		EObject containerOfComposeData = null;
		ComposeData composeData = null;
		if (guiElement.eContainer() instanceof SetOfGuiElements) {
			EObject containerOfSiblings = guiElement.eContainer();
			containerOfComposeData = containerOfSiblings.eContainer().eContainer();
		} else {
			containerOfComposeData = EcoreUtil2.getContainerOfType(guiElement, Composite.class);
		}
		// Todo: create method to getComposeData of a Composite
		if (containerOfComposeData != null) {
			if (containerOfComposeData instanceof IComposite) {
				IComposite iComposite = (IComposite) containerOfComposeData;
				composeData = iComposite.getContent().getComposeLayout();
			} 
		}
		return composeData;
	}

	/**
	 * 
	 * @modified
	 * @return
	 */
	public void addStyle(Style style) {
		StyleProperty styleProperty = (StyleProperty) EcoreUtil.getObjectByType(getProperties(),
				PomPackage.eINSTANCE.getStyleProperty());
		if (null==styleProperty) {
			styleProperty = PomFactory.eINSTANCE.createStyleProperty();
			getProperties().add(styleProperty);
		}
		styleProperty.getStyles().add(style);
	}

	/**
	 * 
	 * @modified
	 * @return
	 */
	public int getHeight() {
		HeightProperty heightProperty = (HeightProperty) EcoreUtil.getObjectByType(getProperties(),
				PomPackage.eINSTANCE.getHeightProperty());
		return DimensionCalculator.calculateHeight(getElementWithLayoutData(),heightProperty);
	}

	/**
	 * 
	 * @modified
	 * @return
	 */
	public int getWidth() {
		WidthProperty widthProperty = (WidthProperty) EcoreUtil.getObjectByType(getProperties(),
			PomPackage.eINSTANCE.getWidthProperty());
		return DimensionCalculator.calculateWidth(getElementWithLayoutData(),widthProperty);		
	}

	/**
	 * 
	 * @modified
	 * @return
	 */
	public void setWidth(int width) {
		WidthProperty widthProperty = (WidthProperty) EcoreUtil.getObjectByType(getProperties(),
				PomPackage.eINSTANCE.getWidthProperty());
		if (null==widthProperty) {
			widthProperty = PomFactory.eINSTANCE.createWidthProperty();
			getProperties().add(widthProperty);
		}
		widthProperty.setWidth(width);
	}
	
	public StyleProperty getStyleProperty() {
		StyleProperty styleProperty = (StyleProperty) EcoreUtil.getObjectByType(getProperties(),
				PomPackage.eINSTANCE.getStyleProperty());		
		return styleProperty;		
	}

	public EList<Style> getStyles() {
		EList<Style> styles = new BasicEList<Style>();
		StyleProperty styleProperty = getStyleProperty();
		if (null != styleProperty) {
			styles = styleProperty.getStyles();
		}
		return styles;
	}

} 
